

package com.ravet.ideaplugins.linemover;


import com.intellij.openapi.actionSystem.DataContext;
import com.intellij.openapi.editor.CaretModel;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.ScrollType;
import com.intellij.openapi.editor.actionSystem.EditorWriteActionHandler;

public final class LineDownMover extends BaseLineMover {

	public LineDownMover() {
		super(new Handler());
	}

	private static class Handler extends EditorWriteActionHandler {

		public void executeWriteAction(Editor editor, DataContext dataContext) {
			if (editor == null) return;

			final Document document = editor.getDocument();
			final CaretModel caretModel = editor.getCaretModel();

			if ((caretModel == null)
			        || (document == null)
			        || (!document.isWritable())
			        || (document.getLineCount() < 2)) {
				return;
			}

			final int lineCount = document.getLineCount();

			final int selectionStart = editor.getSelectionModel().getSelectionStart();
			final int selectionEnd = editor.getSelectionModel().getSelectionEnd();

			final int curLineNumberStart = document.getLineNumber(selectionStart);
			int curLineNumberEnd = document.getLineNumber(selectionEnd);

			if (curLineNumberEnd + 1 >= lineCount) {
				return;
			}

			// adjust end line for a selection ending on
			// column 0 of a multiline selection
			if (curLineNumberStart != curLineNumberEnd) {
				// start of last selected line == end of selection -> last line selected is empty
				if (document.getLineStartOffset(curLineNumberEnd) == selectionEnd) {
				    --curLineNumberEnd;
				}
			}

			final int nextLineNumber = curLineNumberEnd + 1;

			final char[] documentChars = document.getChars();

			final int curBlockStart = document.getLineStartOffset(curLineNumberStart);
			final int curBlockEnd = document.getLineEndOffset(curLineNumberEnd);
			final String curLine = getString(documentChars, curBlockStart, curBlockEnd);

			final int nextBlockStart = document.getLineStartOffset(nextLineNumber);
			int nextBlockEnd = document.getLineEndOffset(nextLineNumber);
			final String nextLine = getString(documentChars, nextBlockStart, nextBlockEnd);


//			debug(selectionStart, selectionEnd, curBlockStart, curBlockEnd,
//			      documentChars, curLine, nextBlockStart, nextBlockEnd, nextLine);


			final int caretOffset = caretModel.getOffset();

			document.replaceString(curBlockStart, nextBlockEnd, nextLine + '\n' + curLine);


			final int selectionCount = selectionEnd - selectionStart;
			final int beginOfSelectionOffset = selectionStart - curBlockStart;

			final int newBeginOfSelectedBlock =
			        curBlockStart + nextLine.length() + 1 + beginOfSelectionOffset;
			final int newEndOfSelectedBlock = newBeginOfSelectedBlock + selectionCount;

			editor.getSelectionModel().setSelection(newBeginOfSelectedBlock, newEndOfSelectedBlock);


			// ensure that two lines below the selection are visible
			final int checkCaretLineNumber =
			        nextLineNumber + 1 < lineCount ? nextLineNumber + 1 : nextLineNumber;
			final int checkCaretOffset = document.getLineStartOffset(checkCaretLineNumber);
			caretModel.moveToOffset(checkCaretOffset);
			editor.getScrollingModel().scrollToCaret(ScrollType.RELATIVE);

			// move the caret if it was within the replaced block
			if (caretOffset >= curBlockStart && caretOffset <= nextBlockEnd) {
				if (caretOffset <= nextBlockStart) {
					final int newCaretOffset = caretOffset + nextBlockEnd - nextBlockStart + 1;
					caretModel.moveToOffset(newCaretOffset);
				}
				else if (caretOffset > nextBlockStart) {
					final int newCaretOffset = curBlockStart + caretOffset - nextBlockStart;
					caretModel.moveToOffset(newCaretOffset);
				}
			}


			editor.getScrollingModel().scrollToCaret(ScrollType.RELATIVE);
		}

		private void debug(int selectionStart, int selectionEnd,
		                   int curBlockStart, int curBlockEnd,
		                   char[] documentChars, String curLine,
		                   int nextBlockStart, int nextBlockEnd,
		                   String nextLine) {

			System.out.println("Line down");
			System.out.println("selectionStart = " + selectionStart);
			System.out.println("selectionEnd = " + selectionEnd);

			System.out.println("curBlockStart = " + curBlockStart);
			System.out.println("curBlockEnd = " + curBlockEnd);
			displayBlock(documentChars, curBlockStart, curBlockEnd);
			displayString(curLine);

			System.out.println("nextBlockStart = " + nextBlockStart);
			System.out.println("nextBlockEnd = " + nextBlockEnd);
			displayBlock(documentChars, nextBlockStart, nextBlockEnd);
			displayString(nextLine);

			System.out.println();
		}

	}
}
