

package com.ravet.ideaplugins.linemover;


import com.intellij.openapi.actionSystem.DataContext;
import com.intellij.openapi.editor.CaretModel;
import com.intellij.openapi.editor.Document;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.editor.ScrollType;
import com.intellij.openapi.editor.actionSystem.EditorWriteActionHandler;

public final class LineUpMover extends BaseLineMover {

	public LineUpMover() {
		super(new Handler());
	}

	private static class Handler extends EditorWriteActionHandler {

		public void executeWriteAction(Editor editor, DataContext dataContext) {
			if (editor == null) return;

			final CaretModel caretModel = editor.getCaretModel();
			final Document document = editor.getDocument();

			if ((caretModel == null)
			        || (document == null)
			        || (!document.isWritable())
			        || (document.getLineCount() < 2)) {
				return;
			}

			final int selectionStart = editor.getSelectionModel().getSelectionStart();
			final int selectionEnd = editor.getSelectionModel().getSelectionEnd();

			final int curLineNumberStart = document.getLineNumber(selectionStart);
			int curLineNumberEnd = document.getLineNumber(selectionEnd);

			final int prevLineNumber = curLineNumberStart - 1;
			if (prevLineNumber < 0) {
				return;
			}

			// adjust end line for a selection ending on
			// column 0 of a multiline selection
			if (curLineNumberStart != curLineNumberEnd) {
				// start of last selected line == end of selection -> last line selected is empty
				if (document.getLineStartOffset(curLineNumberEnd) == selectionEnd) {
				    --curLineNumberEnd;
				}
			}

			final char[] documentChars = document.getChars();

			final int curBlockStart = document.getLineStartOffset(curLineNumberStart);
			final int curBlockEnd = document.getLineEndOffset(curLineNumberEnd);
			final String curLine = getString(documentChars, curBlockStart, curBlockEnd);

			final int prevBlockStart = document.getLineStartOffset(prevLineNumber);
			final int prevBlockEnd = document.getLineEndOffset(prevLineNumber);
			final String prevLine = getString(documentChars, prevBlockStart, prevBlockEnd);


//			debug(selectionStart, selectionEnd, curBlockStart, curBlockEnd,
//			      documentChars, curLine, prevBlockStart, prevBlockEnd, prevLine);


			final int caretOffset = caretModel.getOffset();

			document.replaceString(prevBlockStart, curBlockEnd, curLine + '\n' + prevLine);


			final int selectionOffset = selectionStart - curBlockStart;
			final int selectionCount = selectionEnd - selectionStart;

			final int newBeginOfSelectedBlock = prevBlockStart + selectionOffset;
			final int newEndOfSelectedBlock = newBeginOfSelectedBlock + selectionCount;

			editor.getSelectionModel().setSelection(newBeginOfSelectedBlock, newEndOfSelectedBlock);


			// ensure that two lines above the selection are visible
			final int checkCaretLineNumber = prevLineNumber - 2 >= 0 ? prevLineNumber - 2 : 0;
			final int checkCaretOffset = document.getLineStartOffset(checkCaretLineNumber);
			caretModel.moveToOffset(checkCaretOffset);
			editor.getScrollingModel().scrollToCaret(ScrollType.RELATIVE);

			// move the caret if it was within the replaced block
			if (caretOffset >= prevBlockStart && caretOffset <= curBlockEnd + 1) {
				if (caretOffset > prevBlockEnd) {
					final int newCaretOffset = caretOffset - (prevBlockEnd - prevBlockStart + 1);
					caretModel.moveToOffset(newCaretOffset);
				}
				else if (caretOffset <= prevBlockEnd) {
					final int newCaretOffset = curBlockEnd + caretOffset - prevBlockStart;
					caretModel.moveToOffset(newCaretOffset);
				}
			}

		}

		private void debug(final int selectionStart, final int selectionEnd,
		                   final int curBlockStart, final int curBlockEnd,
		                   final char[] documentChars, final String curLine,
		                   final int prevBlockStart, final int prevBlockEnd,
		                   final String prevLine) {

			System.out.println("Line up");
			System.out.println("selectionStart = " + selectionStart);
			System.out.println("selectionEnd = " + selectionEnd);

			System.out.println("curBlockStart = " + curBlockStart);
			System.out.println("curBlockEnd = " + curBlockEnd);
			displayBlock(documentChars, curBlockStart, curBlockEnd);
			displayString(curLine);

			System.out.println("prevBlockStart = " + prevBlockStart);
			System.out.println("prevBlockEnd = " + prevBlockEnd);
			displayBlock(documentChars, prevBlockStart, prevBlockEnd);
			displayString(prevLine);

			System.out.println();
		}
	}
}